package de.dfki.km.academicresearch.eswc2016challenge.cameraready;

import com.google.common.collect.BiMap;
import com.google.common.collect.HashBiMap;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import org.apache.jena.atlas.lib.Sink;
import org.apache.jena.graph.Triple;
import org.apache.jena.rdf.model.Model;
import org.apache.jena.rdf.model.Statement;
import org.apache.jena.riot.RDFDataMgr;
import org.apache.jena.riot.system.StreamRDF;
import org.apache.jena.riot.system.StreamRDFLib;
import org.nustaq.serialization.FSTObjectInput;
import org.nustaq.serialization.FSTObjectOutput;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 *
 * @author Markus Schr&ouml;der
 */
public class DataBasis {
    
    private static final Logger logger = LoggerFactory.getLogger(DataBasis.class);
    
    private BiMap<String, Integer> uri2id;
    private int[] triplearray;

    private DataBasis() {
        uri2id = HashBiMap.create();
        triplearray = new int[0];
    }
    
    public DataBasis(String filename) throws IOException, ClassNotFoundException {
        FSTObjectInput fstui = new FSTObjectInput(new FileInputStream(new File(filename + ".bimap")));
        uri2id = (BiMap<String, Integer>) fstui.readObject();
        fstui.close();
        
        FSTObjectInput fstta = new FSTObjectInput(new FileInputStream(new File(filename + ".array")));
        triplearray = (int[]) fstta.readObject();
        fstta.close();
    }
    
    public DataBasis(BiMap<String, Integer> uri2id, int[] triplearray) {
        this.uri2id = uri2id;
        this.triplearray = triplearray;
    }

    public DataBasis(Model model) {
        int uriId = 0;
        int tripleId = 0;
        
        int stmtCount = model.listStatements().toSet().size();
        this.uri2id = HashBiMap.create();
        this.triplearray = new int[stmtCount * 3];
        
        for(Statement stmt : model.listStatements().toSet()) {
            
            if(!stmt.getObject().isResource())
                continue;
                
            for(String uri : new String[] { stmt.getSubject().getURI(), stmt.getPredicate().getURI(), stmt.getObject().asResource().getURI() }) {
                if(!uri2id.containsKey(uri)) {
                    uri2id.put(uri, uriId);
                    triplearray[tripleId] = uriId++;
                } else {
                    triplearray[tripleId] = uri2id.get(uri);
                }
                tripleId++;
            }
        }
    }
    
    //==========================================================================
    
    public DataBasis extend(boolean invert, int[] extension) {
        int[] copy = Arrays.copyOf(triplearray, triplearray.length + extension.length);
        if(invert) {
            for(int i = 0; i < triplearray.length; i += 3) {
                int tmp = copy[i+0];
                copy[i+0] = copy[i+2];
                copy[i+2] = tmp;
            }
        }
        for(int i = 0; i < extension.length; i += 3) {
            copy[triplearray.length + i + 0] = extension[i+0];
            copy[triplearray.length + i + 1] = extension[i+1];
            copy[triplearray.length + i + 2] = extension[i+2];
        }
        return new DataBasis(uri2id, copy);
    }
    
    public static DataBasis loadFromNTGZ(String filename) {
        DataBasis data = new DataBasis();
        
        StreamRDF streamRDF = StreamRDFLib.sinkTriples(new Sink<Triple>() {
            
            //BiMap<String, Integer> uri2id = HashBiMap.create();
            int uriId    = 0;
            int tripleId = 0;
            
            
            //33 724 208 * 4 * (32 bits) =
            //539.587328 megabytes
            
            //33 724 208 * 3 * (32 bits) =
            //404.690496 megabytes
            
            //id ist implizit durch die position im array gegeben
            //int[] triplearray = new int[33724208 * 3];
            List<Integer> tripleList = new ArrayList<>();
            
            //um duplicate im eval set gleich auszuschließen
            Set<TripleWrapper> tripleset = new HashSet<>();
            int duplicate = 0;
            
            @Override
            public void send(Triple t) {
                if(t.getObject().isURI()) {
                    
                    //triple
                    int i = 0;
                    Integer[] triple = new Integer[3];
                    
                    //uri to id
                    for(String s : new String[] { t.getSubject().getURI(), t.getPredicate().getURI(), t.getObject().getURI() }) {
                        if(!data.uri2id.containsKey(s)) {
                            data.uri2id.put(s, uriId);
                            triple[i] = uriId;
                            uriId++;
                        } else {
                            triple[i] = data.uri2id.get(s);
                        }
                        i++;
                    }
                    
                    TripleWrapper tw = new TripleWrapper(triple);
                    
                    //id for the triple
                    if(!tripleset.contains(tw)) {
                        tripleId++;
                        tripleset.add(tw);
                        
                        tripleList.add(triple[0]);
                        tripleList.add(triple[1]);
                        tripleList.add(triple[2]);
                    } else {
                        duplicate++;
                    }
                }
            }

            @Override
            public void flush() {
                data.triplearray = toPrimitive(tripleList.toArray(new Integer[0]));
                
                logger.debug("{} uri mappings, {} triples, {} duplicates, ids: {}, {}", data.uri2id.size(), data.triplearray.length, duplicate, uriId, tripleId);
                PrintHeapUsage.printGB();
            }

            @Override
            public void close() {
                
            }
            
        });
        
        RDFDataMgr.parse(streamRDF, filename + ".nt.gz");
        
        return data;
    }
    
    public void save(String filename) {
        try {
            FSTObjectOutput fstoo = new FSTObjectOutput(new FileOutputStream(new File(filename + ".bimap")));
            fstoo.writeObject(uri2id);
            fstoo.close();
        } catch (IOException ex) {
        }

        try {
            FSTObjectOutput fstoo = new FSTObjectOutput(new FileOutputStream(new File(filename + ".array")));
            fstoo.writeObject(triplearray);
            fstoo.close();
        } catch (IOException ex) {
        }
    }
    
    public static int[] toPrimitive(Integer[] IntegerArray) {

		int[] result = new int[IntegerArray.length];
		for (int i = 0; i < IntegerArray.length; i++) {
			result[i] = IntegerArray[i].intValue();
		}
		return result;
	}

    public BiMap<String, Integer> getUri2id() {
        return uri2id;
    }

    public int[] getTriplearray() {
        return triplearray;
    }
    
    public String toString(Integer[] path) {
        return toString(path, true);
    }
    
    public String toString(Integer[] path, boolean min) {
        if(path.length == 0)
            return "";
        
        BiMap<Integer, String> id2uri = uri2id.inverse();
        
        StringBuilder sb = new StringBuilder();
        //first
        if(path[0] == ModifiedEppstein.INIT)
            sb.append("init");
        else
            sb.append(min ? cutURI(id2uri.get(triplearray[path[0]])) : id2uri.get(triplearray[path[0]]));
        sb.append(" ");
        //edges
        for(int edge : path) {
            if(edge == ModifiedEppstein.INIT) {
                sb.append("init <sourceId> ");
            } else {
                for(int r : Arrays.asList(triplearray[edge+1], triplearray[edge+2])) {
                    sb.append(min ? cutURI(id2uri.get(r)) : id2uri.get(r));
                    sb.append(" ");
                }
            }
        }
        return sb.toString();
    }
    
    public String toString(Integer edge) {
        return toString(new Integer[] { edge });
    }
    
    public String toString(List<EdgeDelta> edgedeltas) {
        StringBuilder sb = new StringBuilder();
        for(EdgeDelta ed : edgedeltas) {
            sb.append(toString(ed.edge));
            sb.append(", ");
        }
        return sb.toString();
    }
    
    private String cutURI(String uri) {
        int i1 = uri.lastIndexOf('/');
        if(i1 == -1)
            return uri;
        
        int i2 = uri.lastIndexOf('/', i1 - 1);
        if(i2 == -1)
            return uri;
        
        String a = uri.substring(i2 + 1, i1);
        String b = uri.substring(i1 + 1, uri.length());
        
        if(a.isEmpty())
            return b;
            
        return a.substring(0, 1) + "/" + b;
    }
    
    
    static class TripleWrapper {
        
        Integer[] triple;

        public TripleWrapper(Integer[] triple) {
            this.triple = triple;
        }

        @Override
        public int hashCode() {
            return Arrays.hashCode(triple);
        }

        @Override
        public boolean equals(Object obj) {
            if (this == obj) {
                return true;
            }
            if (obj == null) {
                return false;
            }
            if (getClass() != obj.getClass()) {
                return false;
            }
            final TripleWrapper other = (TripleWrapper) obj;
            if (!Arrays.equals(this.triple, other.triple)) {
                return false;
            }
            return true;
        }
        
    }
    
}
